#!/usr/bin/python
from __future__ import print_function
from boConfig import FTP_LOCAL_PATH, IMAGES
import subprocess
import sys
import os
import json
import time

TIME_STEP = 10 * 60 # the Blitzortung server updates the signal and stroke data every 10 minutes
FOERVER = 10 * 52 * 7 * 24 * 60 * 60 # for the simplicity of the calculation 10 years is the equivalent of 'forever'

logFile = "boGenerateImages.log"

def calcEndOfGenerationTime():
    endTime = int(time.time() / TIME_STEP) * TIME_STEP # now rounded down by TIME_STEP example: now is 08:13:54 then endTime is 08:10:00
    endTime += TIME_STEP # end of the script execution shall be 08:20:00
    endTime -= 15 # keep 15 seconds as reserve for the FTP transfer, end of the image generation shall be at 08:19:45
    return endTime

def parseTime(t = None):
    factor = {
        'y': 52 * 7 * 24 * 60 * 60, # one (simplified) year of 52 weeks
        'w': 7 * 24 * 60 * 60, # one week
        'd': 24 * 60 * 60, # one day
        'h': 60 * 60, # one hour
        'm': 60, # one minute
        's': 1 # one second
        }
    if (None == t):
        seconds = FOERVER
    else:
        val = int(t[0:-1])
        unit = t[-1]
        if unit in factor:
            seconds = val * factor[unit]
        else:
            seconds = FOERVER
    return seconds

def writeDict(file, dict):
    try:
        json.dump(dict, open(file, 'w'))
    except:
        print("could not write file '{}'".format(file))
        e = sys.exc_info()[0]
        print( "Error: {}".format(e))

def readDict(file):
    dict = {}
    try:
        dict = json.load(open(file, 'r'))
    except:
        print("could not read file '{}'".format(file))
        e = sys.exc_info()[0]
        print( "Error: {}".format(e))
    return dict

def cleanupImageInformation(imageInformation, paramsList):
    toBeRemoved = []
    for key in imageInformation:
        if (key not in paramsList):
            toBeRemoved.append(key)
    for key in toBeRemoved:
        del(imageInformation[key])
    return imageInformation

def calcImageTimeCoverage(imageInformation, paramsList):
    for params in paramsList:
        seconds = FOERVER
        for param in params.split():
            if ("-i5" == param): # signals(channel), strokes(channel)/signals(channel), 24h, channel X
                seconds = 24 * 60 * 60
            elif ("-i8" == param): # strokes(channel) and locating ratio 14d
                seconds = 14 * 24 * 60 * 60
            elif ("-i9" == param): # strokes(channel) and locating ratio 24h
                seconds = 24 * 60 * 60
            elif ("-t" == param[0:2]):
                seconds = parseTime(param[2:])
                break
        imageInformation[params]['timeCoverage'] = seconds
    return imageInformation

def calcPriority(priority, timeCoverage):
    if (None == priority):
        priority = 0
        while(timeCoverage > 10):
            timeCoverage /= 10
            priority += 1
    else:
        priority -= 1
    return priority # a small number is urgent, a high number can wait

def calcSortedImageInformation(imageInformation, paramsList):
    for params in paramsList:
        if (params in imageInformation):
            if ('priority' not in imageInformation[params]):
                imageInformation[params]['priority'] = None
            imageInformation[params]['priority'] = calcPriority(imageInformation[params]['priority'], imageInformation[params]['timeCoverage'])
            if ('executionTime' not in imageInformation[params]):
                imageInformation[params]['executionTime'] = 60 # no measured time available, assume 60 seconds required to generate the image
        else:
            print("ERROR: '{}' not found in imageInformation".format(params))
    sortedImageInformation = sorted(imageInformation.iteritems(), key=lambda(x, y): y['priority'])
    return sortedImageInformation

def pickNextImageForGeneration(sortedImageInformation, remainingTime):
    params = None
    
    if len(sortedImageInformation):
        requiredTime = 0
        for i in range(0, len(sortedImageInformation)):
            requiredTime += sortedImageInformation[i][1]['executionTime']
            if (requiredTime > remainingTime):
                break
        if (requiredTime > remainingTime):
            requiredTime -= sortedImageInformation[i][1]['executionTime']
            i -= 1
        if (i >= 0):
            params = sortedImageInformation[i][0]
            #print("requiredTime = {}, remainingTime = {}".format(requiredTime, remainingTime))
            #print("sortedImageInformation[{}] = {}".format(i, sortedImageInformation[i]))
            sortedImageInformation.remove(sortedImageInformation[i])
    return sortedImageInformation, params

def generateImage(sortedImageInformation, imageInformation, params):
    imageGenerated = False
    tStart = time.time()
    cmd = "./boPng.php {}".format(params)
    sp = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = sp.communicate()
    ext = None
    if ("" == err):
        if ("PNG" == out[1:4]):
            ext = "png"
        if ext:
            fileName = "IMG{}.{}".format("".join(i for i in params.split()), ext)
            filePath = "{}/{}".format(FTP_LOCAL_PATH, fileName)
            with open(filePath, "wb") as f:
                f.write(out)
            imageGenerated = True
            executionTime = time.time() - tStart
            imageInformation[params]['fileName'] = fileName # required for the HTML generation
            imageInformation[params]['filePath'] = filePath # required for the HTML generation
            imageInformation[params]['executionTime'] = executionTime # required as information for the next pass
            imageInformation[params]['priority'] = None # the generated image has to get a new priority in the next pass
            imageInformation[params]['generationTime'] = time.time() # information is just informational
            print("{} > {}, {:.1f} sec".format(cmd, fileName, executionTime))
        else:
            print("'{}' failed".format(params))
    else:
        print(err)
    return imageInformation, imageGenerated

def generateImages(sortedImageInformation, imageInformation):
    generatedImageCount = 0
    endTime = calcEndOfGenerationTime()
    remainingTime = endTime - int(time.time())
    while (remainingTime > 0) and len(sortedImageInformation):
        sortedImageInformation, params = pickNextImageForGeneration(sortedImageInformation, remainingTime)
        if (params):
            imageInformation, imageGenerated = generateImage(sortedImageInformation, imageInformation, params)
            if imageGenerated:
                generatedImageCount += 1
            remainingTime = endTime - int(time.time())
        else:
            break
    return imageInformation, generatedImageCount

def generateHtml(imageInformation, paramsList):
    availableImageCount = 0
    html = """<html>
    <head>
    </head>
    <body>
"""
    for params in paramsList:
        if ('fileName' in imageInformation[params]) and ('filePath' in imageInformation[params]):
            if os.path.isfile(imageInformation[params]['filePath']):
                html += "        <img src='{}' />\n".format(imageInformation[params]['fileName'])
                availableImageCount += 1
    html += """    </body>
</html>"""
    file = "{}/test.html".format(FTP_LOCAL_PATH)
    with open(file, "w") as f:
        f.write(html)
    print("{} > {}".format(sys.argv[0], file))
    return availableImageCount

def main():
    paramsList = IMAGES
    if len(sys.argv) > 1:
        paramsList = [''.join("{} ".format(i) for i in sys.argv[1:])]

    imageInformation = readDict(logFile)
    imageInformation = cleanupImageInformation(imageInformation, paramsList)
    imageInformation = calcImageTimeCoverage(imageInformation, paramsList)
    sortedImageInformation = calcSortedImageInformation(imageInformation, paramsList)
    imageInformation, generatedImageCount = generateImages(sortedImageInformation, imageInformation)
    if len(sys.argv) == 1:
        writeDict(logFile, imageInformation)

    availableImageCount = generateHtml(imageInformation, paramsList)
    print("{} images generated, {} images available".format(generatedImageCount, availableImageCount))

if __name__ == "__main__":
    main()
