#!/usr/bin/python
from __future__ import print_function
from boConfig import BO_USER, BO_PASS, BO_STATION, BO_REGION, BOOK_KEEPING_PATH
import time
import calendar
import subprocess
import json
import boDb
import random
import sys
import os
import logging
from geopy.distance import vincenty

desiredExecutionTime = 1 * 60 # note: cron starts this job every 10 minutes, thus some reserve should be there
secStep = 10 * 60 # 10 minutes in seconds
bookKeepingFile = "boGetStrokesFromBlitzortung.log"

url = "http://{}:{}@data.blitzortung.org/Data_{}/Protected/Strokes/%Y/%m/%d/%H/%M.json.gz"

new = 0
done = 0

def configLogger():
    global logger
    format = "%(asctime)-15s %(levelname)s:%(name)s: %(message)s"
    filename = "/var/log/boGetStrokesFromBlitzortung.log"
    logging.basicConfig(filename = filename, format = format)
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    #logger.setLevel(LOGLEVEL)
    boDb.configLogger()

def calcRequiredDownloads():
    """1 download to be done every 10 minutes from 3 servers
    server may not be reachable => retrials may be required
    server may also not have data => no endless retries
    file for bookkeeping
    first line: start and end time/date in human readable format
    second line: end time/date in human readable format
    following lines: downloads to be done from start to end (inclduing borders))
    succeeded lines are deleted
    missing lines for newer dates are to be added
    """
    global secStep
    global bookKeepingFile
    global new

    fileDateTimeFormat = "%Y.%m.%d %H:%M"
    firstLine = True

    urls = []
    secEnd = (int(time.time() / secStep) * secStep) - secStep # multiple of secStep, older than now
    try:
        with open(bookKeepingFile, 'rt') as f:
            for line in f:
                line = line.rstrip()
                if (line):
                    urls.append(line)
                if (firstLine):
                    firstLine = False
                    line = line.split('-')
                    tInitialStart = time.strptime(line[0].strip(), fileDateTimeFormat)
                    secInitialStart = int(calendar.timegm(tInitialStart) / secStep) * secStep
                    tInitialEnd = time.strptime(line[1].strip(), fileDateTimeFormat)
                    secInitialEnd = int(calendar.timegm(tInitialEnd) / secStep) * secStep
                    secStart = secInitialEnd + secStep
    except:
        secInitialStart = secEnd - 3 * 60 * 60 # signals are available for the previous 3 hours, thus initially start with this time range
        secStart = secInitialStart
        urls.append("")
    urls[0] = time.strftime(fileDateTimeFormat, time.gmtime(secInitialStart)) + " - " + time.strftime(fileDateTimeFormat, time.gmtime(secEnd))

    new = 0
    for sec in range(secStart, secEnd + secStep, secStep):
        t = time.gmtime(sec)
        for i in BO_REGION:
            link = url.format(BO_USER, BO_PASS, i)
            link = time.strftime(link, t)
            urls.append(link)
            new += 1
    return urls

def savePendingDownloads(urls):
    global new
    with open(bookKeepingFile, 'wt') as f:
        for line in urls:
            f.write(line + '\n')
    logger.info("{} downloaded: {} new, {} done, {} pending".format(urls[0], new, done, len(urls) - 1))

def readStderr(sp, cmd):
    err = ""
    line = sp.stderr.readline()
    while line:
        err += "{}: {}".format(cmd, line)
        line = sp.stderr.readline()
    return err

def getStrokeStationData(line):
    stroke = {}
    j = {}
    try:
        j = json.loads(line)
    except:
        logger.error("parsing json data failed: '{}'".format(sys.exc_info()[0]))
    if ("sig" in j):
        for i in reversed(range(len(j["sig"]))):
            if (BO_STATION == j["sig"][i]["sta"]):
                stroke["time_ns_stroke"] = j["time"]
                stroke["lat_stroke"] = j["lat"]
                stroke["lon_stroke"] = j["lon"]
                stroke["alt_stroke"] = j["alt"]
                stroke["delta_time_ns_station"] = j["sig"][i]["time"]
                stroke["lat_station"] = j["sig"][i]["lat"]
                stroke["lon_station"] = j["sig"][i]["lon"]
                stroke["alt_station"] = j["sig"][i]["alt"]
    else:
        logger.error("no 'sig' in the data line '{}'".format(line))
    return stroke

def calcDistance(stroke):
    a = (stroke["lat_stroke"], stroke["lon_stroke"])
    b = (stroke["lat_station"], stroke["lon_station"])
    distance = vincenty(a, b).km
    return round(distance, 1)

def urlGetGrepParse(url):
    result = False

    cmd1 = "wget -qO- {}".format(url).split()
    sp1 = subprocess.Popen(cmd1, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    cmd2 = "grep \"sta\":\"{}\"".format(BO_STATION).split()
    sp2 = subprocess.Popen(cmd2, stdin=sp1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    strokeCnt = 0
    strokeInsertFailed = 0
    line = sp2.stdout.readline()
    while line:
        stroke = getStrokeStationData(line)
        if (stroke):
            stroke["distance"] = calcDistance(stroke)
            if not boDb.insertStroke(stroke):
                strokeInsertFailed += 1
            strokeCnt += 1
        line = sp2.stdout.readline()

    err = readStderr(sp1, "wget")
    err += readStderr(sp2, "grep")

    if ("" == err):
        if (strokeCnt or strokeInsertFailed):
            if (strokeInsertFailed):
                logger.info("'{}' contained {} strokes, insert failed for {}".format(url, strokeCnt, strokeInsertFailed))
            else:
                logger.info("'{}' contained {} strokes".format(url, strokeCnt))
        return True
    else:
        logger.debug(err)
        return False

def urlGetUnzipGrepParse(url):
    result = False

    cmd1 = "wget -qO- {}".format(url).split()
    sp1 = subprocess.Popen(cmd1, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    cmd2 = "gunzip -c".split()
    sp2 = subprocess.Popen(cmd2, stdin=sp1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    cmd3 = "grep \"sta\":\"{}\"".format(BO_STATION).split()
    sp3 = subprocess.Popen(cmd3, stdin=sp2.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    strokeCnt = 0
    strokeInsertFailed = 0
    line = sp3.stdout.readline()
    while line:
        stroke = getStrokeStationData(line)
        if (stroke):
            stroke["distance"] = calcDistance(stroke)
            if not boDb.insertStroke(stroke):
                strokeInsertFailed += 1
            strokeCnt += 1
        line = sp3.stdout.readline()

    err = readStderr(sp1, "wget")
    err += readStderr(sp2, "gunzip")
    err += readStderr(sp3, "grep")

    if ("" == err):
        if (strokeCnt or strokeInsertFailed):
            if (strokeInsertFailed):
                logger.info("'{}' contained {} strokes, insert failed for {}".format(url, strokeCnt, strokeInsertFailed))
            else:
                logger.info("'{}' contained {} strokes".format(url, strokeCnt))
        return True
    else:
        logger.debug(err)
        return False

def correlateSignalsStrokes():
    existingLinks = boDb.querySignalsStrokes()
    links = boDb.correlateSignalsStrokes()
    links += boDb.correlateSignalsStrokes(-999999999)
    linkCnt = 0
    linkUpdateFailed = 0
    for link in links:
        if (link not in existingLinks):
            if not boDb.updateStrokeSignal(link):
                linkUpdateFailed += 1
            linkCnt += 1
    if (linkUpdateFailed):
        logger.info("correlating signals and strokes: {} links identified, update failed for {}".format(linkCnt, linkUpdateFailed))
    else:
        logger.info("correlating signals and strokes: {} links identified".format(linkCnt))

def getBookKeepingPath():
    global bookKeepingFile
    if ("" == BOOK_KEEPING_PATH):
        return
    if (os.path.exists(BOOK_KEEPING_PATH) and os.path.isdir(BOOK_KEEPING_PATH)):
        if ('/' == BOOK_KEEPING_PATH[-1]):
            bookKeepingFile = BOOK_KEEPING_PATH + bookKeepingFile
        else:
            bookKeepingFile = BOOK_KEEPING_PATH + '/' + bookKeepingFile
    else:
        logger.error("BOOK_KEEPING_PATH = '{}' does not exist".format(BOOK_KEEPING_PATH))
        sys.exit(1)

def main():
    # normal program starts here
    global done
    global desiredExecutionTime
    done = 0
    configLogger()
    tEnd = time.time() + desiredExecutionTime
    random.seed()
    getBookKeepingPath()
    boDb.connectMysql()
    boDb.connectDatabase()
    urls = calcRequiredDownloads()
    time.sleep(30) # allow the server to finalize the file - dealy due to fear to miss something not due to knowledge
    while (tEnd > time.time()):
        count = len(urls) - 1
        if (0 == count):
            break
        i = random.randint(1, count)
        url = urls[i]
        result = urlGetUnzipGrepParse(url) # try the .gz version first in order to save bandwidth
        if (True != result):
            url = url[0:-3] # if failed, remove the '.gz' from the url
            result = urlGetGrepParse(url)
        if (True == result):
            del urls[i]
            done += 1
        pass
    savePendingDownloads(urls)
    correlateSignalsStrokes()
    boDb.disconnectMysql()

if __name__ == "__main__":
    main()
